"""Implementation of sanity rule."""

from __future__ import annotations

import re
import sys
from typing import TYPE_CHECKING

from ansiblelint.rules import AnsibleLintRule

# Copyright (c) 2018, Ansible Project


if TYPE_CHECKING:
    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable


class CheckSanityIgnoreFiles(AnsibleLintRule):
    """Ignore entries in sanity ignore files must match an allow list."""

    id = "sanity"
    description = (
        "Identifies non-allowed entries in the `tests/sanity/ignore*.txt files."
    )
    severity = "MEDIUM"
    tags = ["idiom"]
    version_changed = "6.14.0"

    # Partner Engineering defines this list. Please contact PE for changes.

    allowed_ignores = [
        "validate-modules:missing-gplv3-license",
        "action-plugin-docs",  # Added for Networking Collections
        "import-2.6",
        "import-2.6!skip",
        "import-2.7",
        "import-2.7!skip",
        "import-3.5",
        "import-3.5!skip",
        "compile-2.6",
        "compile-2.6!skip",
        "compile-2.7",
        "compile-2.7!skip",
        "compile-3.5",
        "compile-3.5!skip",
        "shebang",  # Unreliable test
        "shellcheck",  # Unreliable test
        "pylint:used-before-assignment",  # Unreliable test
    ]

    no_check_ignore_files = [
        "ignore-2.9",
        "ignore-2.10",
        "ignore-2.11",
        "ignore-2.12",
    ]

    _ids = {
        "sanity[cannot-ignore]": "Ignore file contains ... at line ..., which is not a permitted ignore.",
        "sanity[bad-ignore]": "Ignore file entry at ... is formatted incorrectly. Please review.",
    }

    def matchyaml(self, file: Lintable) -> list[MatchError]:
        """Evaluate sanity ignore lists for disallowed ignores.

        :param file: Input lintable file that is a match for `sanity-ignore-file`
        :returns: List of errors matched to the input file
        """
        results: list[MatchError] = []
        test = ""

        check_dirs = {
            "plugins",
            "roles",
        }

        if file.kind != "sanity-ignore-file":
            return []

        with file.path.open(encoding="utf-8") as ignore_file:
            entries = ignore_file.read().splitlines()

            if any(name in str(file.abspath) for name in self.no_check_ignore_files):
                return []

            for line_num, entry in enumerate(entries, 1):
                base_ignore_dir = ""

                if entry:
                    # match up to the first "/"
                    regex = re.match(r"[^/]*", entry)

                    if regex:
                        base_ignore_dir = regex.group(0)

                    if base_ignore_dir in check_dirs:
                        try:
                            if "#" in entry:
                                entry, _ = entry.split("#", 1)
                            (_, test) = entry.split()
                            if test not in self.allowed_ignores:
                                results.append(
                                    self.create_matcherror(
                                        message=f"Ignore file contains {test} at line {line_num}, which is not a permitted ignore.",
                                        tag="sanity[cannot-ignore]",
                                        lineno=line_num,
                                        filename=file,
                                    ),
                                )

                        except ValueError:
                            results.append(
                                self.create_matcherror(
                                    message=f"Ignore file entry at {line_num} is formatted incorrectly. Please review.",
                                    tag="sanity[bad-ignore]",
                                    lineno=line_num,
                                    filename=file,
                                ),
                            )

        return results


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("test_file", "failures", "tags"),
        (
            pytest.param(
                "examples/sanity_ignores/tests/sanity/ignore-2.9.txt",
                0,
                "sanity[cannot-ignore]",
                id="pass",
            ),
            pytest.param(
                "examples/sanity_ignores/tests/sanity/ignore-2.15.txt",
                1,
                "sanity[bad-ignore]",
                id="fail0",
            ),
            pytest.param(
                "examples/sanity_ignores/tests/sanity/ignore-2.13.txt",
                1,
                "sanity[cannot-ignore]",
                id="fail1",
            ),
        ),
    )
    def test_sanity_ignore_files(
        default_rules_collection: RulesCollection,
        test_file: str,
        failures: int,
        tags: str,
    ) -> None:
        """Test rule matches."""
        default_rules_collection.register(CheckSanityIgnoreFiles())
        results = Runner(test_file, rules=default_rules_collection).run()
        for result in results:
            assert result.rule.id == CheckSanityIgnoreFiles().id
            assert result.tag == tags
        assert len(results) == failures



"""Module containing cached JSON schemas."""

import argparse
import json
import logging
import os
import sys
import time
import urllib.request
from collections import defaultdict
from functools import cache
from http.client import HTTPException
from pathlib import Path
from typing import Any
from urllib.error import HTTPError
from urllib.request import Request

_logger = logging.getLogger(__package__)

# Maps kinds to JSON schemas
# See https://www.schemastore.org/json/
store_file = Path(f"{__file__}/../__store__.json").resolve()
with store_file.open(encoding="utf-8") as json_file:
    JSON_SCHEMAS = json.load(json_file)


class SchemaCacheDict(defaultdict):  # type: ignore[type-arg]
    """Caching schema store."""

    def __missing__(self, key: str) -> Any:
        """Load schema on its first use."""
        value = get_schema(key)
        self[key] = value
        return value


@cache
def get_schema(kind: str) -> Any:
    """Return the schema for the given kind."""
    schema_file = Path(__file__).parent / f"{kind}.json"
    with schema_file.open(encoding="utf-8") as f:
        return json.load(f)


_schema_cache = SchemaCacheDict()


def _fetch_and_store_schema(
    request: Request,
    kind: str,
    path: Path,
    data: dict[str, Any],
    schema: dict[str, Any],
) -> int:
    """Fetch a schema from URL and save it to disk. Returns 1 if changed, 0 otherwise."""
    with urllib.request.urlopen(request, timeout=10) as response:  # noqa: S310
        if response.status == 200:  # pragma: no cover
            content = response.read().decode("utf-8").rstrip()
            etag = response.headers["etag"].strip('"')
            changed = 0
            if etag != data.get("etag", "") or not schema:
                JSON_SCHEMAS[kind]["etag"] = etag
                changed = 1
            with path.open("w", encoding="utf-8") as f_out:
                _logger.info("Schema %s was updated", kind)
                f_out.write(content)
                f_out.write("\n")  # prettier/editors
                f_out.truncate()
                os.fsync(f_out.fileno())
                if kind in _schema_cache:  # pragma: no cover
                    del _schema_cache[kind]
            return changed
    return 0  # pragma: no cover


def refresh_schemas(min_age_seconds: int = 3600 * 24) -> int:
    """Refresh JSON schemas by downloading latest versions.

    Returns number of changed schemas.
    """
    age = int(time.time() - store_file.stat().st_mtime)

    # never check for updated schemas more than once a day
    if min_age_seconds > age:
        return 0
    if not os.access(store_file, os.W_OK):  # pragma: no cover
        _logger.debug(
            "Skipping schema update due to lack of writing rights on %s",
            store_file,
        )
        return -1
    _logger.debug("Checking for updated schemas...")

    changed = 0
    for kind, data in JSON_SCHEMAS.items():
        try:
            schema = get_schema(kind)
        except OSError:
            _logger.exception(
                "Error getting schema %s", kind, stack_info=False, stacklevel=2
            )
            schema = {}
        url = data["url"]
        if "#" in url:  # pragma: no cover
            msg = f"Schema URLs cannot contain # due to python-jsonschema limitation: {url}"
            raise RuntimeError(msg)
        path = Path(__file__).parent.resolve() / f"{kind}.json"
        _logger.debug("Refreshing %s schema ...", kind)
        if not url.startswith(("http:", "https:")):  # pragma: no cover
            msg = f"Unexpected url schema: {url}"
            raise ValueError(msg)
        request = Request(url)  # noqa: S310
        etag = data.get("etag", "")
        if not path.exists():
            etag = ""
        if etag:
            request.add_header("If-None-Match", f'"{data.get("etag")}"')
        try:
            changed += _fetch_and_store_schema(request, kind, path, data, schema)
        except (ConnectionError, OSError, HTTPException) as exc:
            if isinstance(exc, HTTPError) and getattr(exc, "code", None) == 304:
                _logger.debug("Schema %s is not modified", url)
                continue
            # In case of networking issues, we just stop and use last-known good
            _logger.debug("Skipped schema refresh due to unexpected exception: %s", exc)
            break
    if changed:  # pragma: no cover
        with store_file.open("w", encoding="utf-8") as f_out:
            # formatting should match our .prettierrc.yaml
            json.dump(JSON_SCHEMAS, f_out, indent=2, sort_keys=True)
            f_out.write("\n")  # prettier and editors in general
        # clear schema cache
        get_schema.cache_clear()
    else:
        store_file.touch()
    return changed


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--force", "-f", action="store_true")
    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)

    if refresh_schemas(0 if args.force else 60 * 10):  # pragma: no cover
        print("Schemas were updated.")  # noqa: T201
        sys.exit(1)
    else:  # pragma: no cover
        print("Schemas not updated", 0)  # noqa: T201



"""Implementation for deprecated-local-action rule."""

# Copyright (c) 2016, Tsukinowa Inc. <info@tsukinowa.jp>
# Copyright (c) 2018, Ansible Project
from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

from ruamel.yaml.comments import CommentedMap

from ansiblelint.errors import MatchError
from ansiblelint.file_utils import Lintable
from ansiblelint.rules import AnsibleLintRule, TransformMixin
from ansiblelint.runner import get_matches
from ansiblelint.transformer import Transformer

if TYPE_CHECKING:
    from ruamel.yaml.comments import CommentedSeq

    from ansiblelint.config import Options
    from ansiblelint.utils import Task


_logger = logging.getLogger(__name__)


class LocalActionTransformError(Exception):
    """Exception raised when a local_action is not processed correctly."""

    def __init__(self, message: str) -> None:
        """Initialize the LocalActionTransformError with a message."""
        _logger.error(message)


class TaskNoLocalActionRule(AnsibleLintRule, TransformMixin):
    """Do not use 'local_action', use 'delegate_to: localhost'."""

    id = "deprecated-local-action"
    description = "Do not use ``local_action``, use ``delegate_to: localhost``"
    needs_raw_task = True
    severity = "MEDIUM"
    tags = ["deprecations"]
    version_changed = "4.0.0"

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> bool | str:
        """Return matches for a task."""
        raw_task = task["__raw_task__"]
        return "local_action" in raw_task

    def transform(
        self,
        match: MatchError,
        lintable: Lintable,
        data: CommentedMap | CommentedSeq | str,
    ) -> None:
        """Transform the task to use delegate_to: localhost.

        Args:
            match: The match object.
            lintable: The lintable object.
            data: The data to transform.
        """
        try:
            self.perform_transform(match, lintable, data)
        except LocalActionTransformError as e:
            match.fixed = False
            match.message = str(e)
            return

    def perform_transform(
        self,
        match: MatchError,
        lintable: Lintable,
        data: CommentedMap | CommentedSeq | str,
    ) -> None:
        """Transform the task to use delegate_to: localhost.

        Args:
            match: The match object.
            lintable: The lintable object.
            data: The data to transform.

        Raises:
            LocalActionTransformError: If the local_action is not dict | str.
        """
        original_task = self.seek(match.yaml_path, data)
        task_location = f"{lintable.name}:{match.lineno}"

        target_task = {}

        for k, v in original_task.items():
            if k == "local_action":
                if isinstance(v, dict):
                    target_task.update(self.process_dict(v, task_location))
                elif isinstance(v, str):
                    target_task.update(self.process_string(v, task_location))
                else:
                    err = f"Unsupported local_action type '{type(v).__name__}' in task at {task_location}"
                    raise LocalActionTransformError(err)
                target_task["delegate_to"] = "localhost"
            else:
                target_task[k] = v

        match.fixed = True
        original_task.clear()
        original_task.update(target_task)

    def process_dict(
        self, local_action: dict[str, Any], task_location: str
    ) -> dict[str, Any]:
        """Process a dict-form local_action.

        Args:
            local_action: The local_action dictionary.
            task_location: The location of the task.

        Returns:
            A dictionary with the module and parameters.

        Raises:
            LocalActionTransformError: If the local_action dictionary is missing a 'module' key.
        """
        if "module" not in local_action:
            err = f"No 'module' key in local_action in task at {task_location}"
            raise LocalActionTransformError(err)
        return {
            local_action["module"]: {
                k: v for k, v in local_action.items() if k != "module"
            }
            or None
        }

    def process_string(
        self, local_action: str, task_location: str
    ) -> dict[str, str | None]:
        """Process a string-form local_action.

        Args:
            local_action: The local_action string.
            task_location: The location of the task.

        Returns:
            A dictionary with the module and parameters.

        Raises:
            LocalActionTransformError: If the local_action string is empty or whitespace.
        """
        if not local_action or not local_action.strip():
            err = f"Empty local_action string in task at {task_location}"
            raise LocalActionTransformError(err)
        parts = local_action.split(" ", 1)
        return {parts[0]: parts[1] if len(parts) > 1 else None}


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    def test_local_action(default_rules_collection: RulesCollection) -> None:
        """Positive test deprecated_local_action."""
        results = Runner(
            "examples/playbooks/rule-deprecated-local-action-fail.yml",
            rules=default_rules_collection,
        ).run()

        assert any(result.tag == "deprecated-local-action" for result in results)

    @pytest.mark.parametrize(
        ("data", "prefix"),
        (
            (
                CommentedMap({"local_action": True}),
                "Unsupported local_action type 'bool'",
            ),
            (
                CommentedMap({"local_action": 123}),
                "Unsupported local_action type 'int'",
            ),
            (
                CommentedMap({"local_action": 12.34}),
                "Unsupported local_action type 'float'",
            ),
            (
                CommentedMap({"local_action": []}),
                "Unsupported local_action type 'list'",
            ),
            (CommentedMap({"local_action": {}}), "No 'module' key in local_action"),
            (CommentedMap({"local_action": ""}), "Empty local_action string"),
            (CommentedMap({"local_action": "  "}), "Empty local_action string"),
        ),
        ids=[
            "bool",
            "int",
            "float",
            "list",
            "empty_dict",
            "empty_string",
            "whitespace_string",
        ],
    )
    def test_local_action_transform_fail(
        caplog: pytest.LogCaptureFixture, data: CommentedMap, prefix: str
    ) -> None:
        """Test transform functionality for a failure.

        Args:
            caplog: The pytest fixture for capturing logs.
            data: The data to test.
            prefix: The expected error prefix.
        """
        file = "site.yml"
        rule = TaskNoLocalActionRule()
        lintable = Lintable(name=file)
        match_error = MatchError(message="error", lintable=lintable, lineno=1)
        with pytest.raises(LocalActionTransformError):
            rule.perform_transform(match_error, lintable, data)
        assert f"{prefix} in task at {file}:1" in caplog.text

    def test_local_action_transform(
        config_options: Options,
        monkeypatch: pytest.MonkeyPatch,
        empty_rule_collection: RulesCollection,
    ) -> None:
        """Test transform functionality for no-log-password rule."""
        monkeypatch.setenv("ANSIBLE_LINT_WRITE_TMP", "1")

        playbook = Path("examples/playbooks/tasks/local_action.yml")
        config_options.write_list = ["all"]

        config_options.lintables = [str(playbook)]
        empty_rule_collection.register(TaskNoLocalActionRule())
        runner_result = get_matches(
            rules=empty_rule_collection,
            options=config_options,
        )
        transformer = Transformer(result=runner_result, options=config_options)
        transformer.run()
        matches = runner_result.matches
        assert any(error.tag == "deprecated-local-action" for error in matches)

        orig_content = playbook.read_text(encoding="utf-8")
        expected_content = playbook.with_suffix(
            f".transformed{playbook.suffix}",
        ).read_text(encoding="utf-8")
        transformed_content = playbook.with_suffix(f".tmp{playbook.suffix}").read_text(
            encoding="utf-8",
        )

        assert orig_content != transformed_content
        assert expected_content == transformed_content

        playbook.with_suffix(f".tmp{playbook.suffix}").unlink()



"""Implementation of limiting number of tasks."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

from ansiblelint.rules import AnsibleLintRule, RulesCollection

if TYPE_CHECKING:
    from ansiblelint.app import App
    from ansiblelint.config import Options
    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class ComplexityRule(AnsibleLintRule):
    """Sets maximum complexity to avoid complex plays."""

    id = "complexity"
    description = "Checks for complex plays and tasks"
    link = "https://ansible.readthedocs.io/projects/lint/rules/complexity/"
    severity = "MEDIUM"
    tags = ["experimental"]
    version_changed = "26.2.0"

    def __init__(self) -> None:
        """Initialize the rule."""
        super().__init__()
        self._collection: RulesCollection | None = None

    def matchplay(self, file: Lintable, data: dict[str, Any]) -> list[MatchError]:
        """Call matchplay for up to no_of_max_tasks inside file and return aggregate results."""
        results: list[MatchError] = []

        if file.kind != "playbook":
            return []
        tasks = data.get("tasks", [])
        if not isinstance(self._collection, RulesCollection):  # pragma: no cover
            msg = "Rules cannot be run outside a rule collection."
            raise TypeError(msg)
        if len(tasks) > self._collection.options.max_tasks:
            results.append(
                self.create_matcherror(
                    message=f"Maximum tasks allowed in a play is {self._collection.options.max_tasks}.",
                    tag=f"{self.id}[play]",
                    filename=file,
                    data=data,
                ),
            )
        return results

    def matchtask(self, task: Task, file: Lintable | None = None) -> list[MatchError]:
        """Check if the task is a block and count the number of items inside it."""
        results: list[MatchError] = []

        if not isinstance(self._collection, RulesCollection):  # pragma: no cover
            msg = "Rules cannot be run outside a rule collection."
            raise TypeError(msg)

        if task.action == "block/always/rescue":
            block_depth = self.calculate_block_depth(task)
            if block_depth > self._collection.options.max_block_depth:
                results.append(
                    self.create_matcherror(
                        message=f"Replace nested block with an include_tasks to make code easier to maintain. Maximum block depth allowed is {self._collection.options.max_block_depth}.",
                        lineno=task.line,
                        tag=f"{self.id}[nesting]",
                        filename=file,
                    ),
                )
        return results

    def matchtasks(self, file: Lintable) -> list[MatchError]:
        """Call matchtask for each task and check total task count."""
        matches: list[MatchError] = []

        if not isinstance(self._collection, RulesCollection):  # pragma: no cover
            msg = "Rules cannot be run outside a rule collection."
            raise TypeError(msg)

        # Call parent's matchtasks to get all individual task violations
        matches = super().matchtasks(file)

        # Only check total task count for task files and handler files
        # Playbooks use the complexity[play] check instead
        if file.kind in ["handlers", "tasks"]:
            # pylint: disable=import-outside-toplevel
            from ansiblelint.utils import task_in_list

            task_count = sum(
                1
                for _ in task_in_list(
                    data=file.data,
                    file=file,
                    kind=file.kind,
                )
            )

            # Check if total task count exceeds limit
            if task_count > self._collection.options.max_tasks:
                matches.append(
                    self.create_matcherror(
                        message=f"File contains {task_count} tasks, exceeding the maximum of {self._collection.options.max_tasks}. Consider using `ansible.builtin.include_tasks` to split the tasks into smaller files.",
                        tag=f"{self.id}[tasks]",
                        filename=file,
                    ),
                )

        return matches

    def calculate_block_depth(self, task: Task) -> int:
        """Recursively calculate the block depth of a task."""
        if not isinstance(task.position, str):  # pragma: no cover
            raise NotImplementedError
        return task.position.count(".block")


if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("file", "expected_results"),
        (
            pytest.param(
                "examples/playbooks/rule-complexity-pass.yml",
                [],
                id="pass",
            ),
            pytest.param(
                "examples/playbooks/rule-complexity-fail.yml",
                ["complexity[play]", "complexity[nesting]"],
                id="fail",
            ),
            pytest.param(
                "examples/playbooks/tasks/rule-complexity-tasks-fail.yml",
                ["complexity[tasks]"],
                id="tasks",
            ),
        ),
    )
    def test_complexity(
        file: str,
        expected_results: list[str],
        monkeypatch: pytest.MonkeyPatch,
        config_options: Options,
        app: App,
    ) -> None:
        """Test rule."""
        monkeypatch.setattr(config_options, "max_tasks", 5)
        monkeypatch.setattr(config_options, "max_block_depth", 3)
        collection = RulesCollection(app=app, options=config_options)
        collection.register(ComplexityRule())
        results = Runner(file, rules=collection).run()

        assert len(results) == len(expected_results)
        for i, result in enumerate(results):
            assert result.rule.id == ComplexityRule.id, result
            assert result.tag == expected_results[i]



"""Exceptions and error representations."""

from __future__ import annotations

import functools
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from ansiblelint._internal.rules import BaseRule, RuntimeErrorRule
from ansiblelint.file_utils import Lintable

if TYPE_CHECKING:
    from ansiblelint.utils import Task


class LintWarning(Warning):
    """Used by linter."""


@dataclass
class WarnSource:
    """Container for warning information, so we can later create a MatchError from it."""

    filename: Lintable
    lineno: int
    tag: str
    message: str | None = None


@dataclass(frozen=True)
class RuleMatchTransformMeta:
    """Additional metadata about a match error to be used during transformation."""


# pylint: disable=too-many-instance-attributes
@dataclass(unsafe_hash=True)
@functools.total_ordering
class MatchError(ValueError):
    """Rule violation detected during linting.

    It can be raised as Exception but also just added to the list of found
    rules violations.

    Note that line argument is not considered when building hash of an
    instance.
    """

    # order matters for these:
    message: str = field(init=True, repr=False, default="")
    lintable: Lintable = field(init=True, repr=False, default=Lintable(name=""))

    tag: str = field(init=True, repr=False, default="")
    lineno: int = 1
    details: str = ""
    column: int | None = None
    # rule is not included in hash because we might have different instances
    # of the same rule, but we use the 'tag' to identify the rule.
    rule: BaseRule = field(hash=False, default=RuntimeErrorRule())
    ignored: bool = False
    fixed: bool = False  # True when a transform has resolved this MatchError
    transform_meta: RuleMatchTransformMeta | None = None

    def __post_init__(self) -> None:
        """Can be use by rules that can report multiple errors type, so we can still filter by them."""
        self.filename = self.lintable.name

        # We want to catch accidental MatchError() which contains no useful
        # information. When no arguments are passed, the '_message' field is
        # set to 'property', only if passed it becomes a string.
        if self.rule.__class__ is RuntimeErrorRule:
            # so instance was created without a rule
            if not self.message:
                msg = f"{self.__class__.__name__}() missing a required argument: one of 'message' or 'rule'"
                raise TypeError(msg)
            if not isinstance(self.tag, str):  # pragma: no cover
                msg = "MatchErrors must be created with either rule or tag specified."
                raise TypeError(msg)
        if not self.message:
            self.message = self.rule.shortdesc

        self.match_type: str | None = None
        # for task matches, save the normalized task object (useful for transforms)
        self.task: Task | None = None
        # path to the problem area, like: [0,"pre_tasks",3] for [0].pre_tasks[3]
        self.yaml_path: list[int | str] = []

        if not self.tag:
            self.tag = self.rule.id

        # Safety measure to ensure we do not end-up with incorrect indexes
        if self.lineno == 0:  # pragma: no cover
            msg = "MatchError called incorrectly as line numbers start with 1"
            raise RuntimeError(msg)
        if self.column == 0:  # pragma: no cover
            msg = "MatchError called incorrectly as column numbers start with 1"
            raise RuntimeError(msg)

        self.lineno += self.lintable.line_offset

        # We make the lintable aware that we found a match inside it, as this
        # can be used to skip running other rules that do require current one
        # to pass.
        self.lintable.matches.append(self)

    @functools.cached_property
    def level(self) -> str:
        """Return the level of the rule: error, warning or notice."""
        if (
            not self.ignored
            and self.rule.options
            and {self.tag, self.rule.id, *self.rule.tags}.isdisjoint(
                self.rule.options.warn_list,
            )
        ):
            return "error"
        return "warning"

    def __repr__(self) -> str:
        """Return a MatchError instance representation."""
        formatstr = "[{0}] ({1}) matched {2}:{3} {4}"
        # note that `rule.id` can be int, str or even missing, as users
        # can defined their own custom rules.
        id_ = getattr(self.rule, "id", "000")

        return formatstr.format(
            id_,
            self.message,
            self.filename,
            self.lineno,
            self.details,
        )

    def __str__(self) -> str:
        """Return a MatchError instance string representation."""
        return self.__repr__()

    @property
    def position(self) -> str:
        """Return error positioning, with column number if available."""
        if self.column:
            return f"{self.lineno}:{self.column}"
        return str(self.lineno)

    @property
    def _hash_key(self) -> Any:
        # line attr is knowingly excluded, as dict is not hashable
        return (
            self.filename,
            self.lineno,
            str(getattr(self.rule, "id", 0)),
            self.message,
            self.details,
            # -1 is used here to force errors with no column to sort before
            # all other errors.
            -1 if self.column is None else self.column,
        )

    def __lt__(self, other: object) -> bool:
        """Return whether the current object is less than the other."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return bool(self._hash_key < other._hash_key)

    def __eq__(self, other: object) -> bool:
        """Identify whether the other object represents the same rule match."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.__hash__() == other.__hash__()



"""Utility functions related to file operations."""

from __future__ import annotations

import copy
import logging
import os
import sys
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, cast

import pathspec
import wcmatch.pathlib
from yaml.error import YAMLError

from ansiblelint.config import ANSIBLE_OWNED_KINDS, BASE_KINDS, Options, options
from ansiblelint.constants import CONFIG_FILENAMES, FileType, States

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from ansiblelint.errors import MatchError


_logger = logging.getLogger(__package__)

ROLE_SUBDIRS = ("tasks", "meta", "vars", "defaults", "handlers")


def _has_role_subdirs(path: Path) -> bool:
    """Return true if path is a directory with typical role subfolders."""
    return path.is_dir() and any((path / subdir).is_dir() for subdir in ROLE_SUBDIRS)


def find_role_dir(path: Path) -> Path | None:
    """Return the innermost roles/ descendant that looks like a role root."""
    current = path if path.is_dir() else path.parent
    found: Path | None = None
    while current.name:
        if (
            "roles" in current.parts
            and current.name != "roles"
            and _has_role_subdirs(current)
        ):
            found = current
        current = current.parent
    return found


def abspath(path: str, base_dir: str) -> str:
    """Make relative path absolute relative to given directory.

    path (str): the path to make absolute
    base_dir (str): the directory from which make relative paths absolute.
    """
    if not os.path.isabs(path):
        # Don't use abspath as it assumes path is relative to cwd.
        # We want it relative to base_dir.
        path = os.path.join(base_dir, path)

    return os.path.normpath(path)


def normpath(path: str | Path) -> str:
    """Normalize a path in order to provide a more consistent output.

    Currently it generates a relative path but in the future we may want to
    make this user configurable.
    """
    # prevent possible ValueError with relpath(), when input is an empty string
    if not path:
        path = "."
    # conversion to string in order to allow receiving non string objects
    relpath = os.path.relpath(str(path))
    path_absolute = os.path.abspath(str(path))
    if path_absolute.startswith(os.getcwd()):
        return relpath
    if path_absolute.startswith(os.path.expanduser("~")):
        return path_absolute.replace(os.path.expanduser("~"), "~")
    # we avoid returning relative paths that end-up at root level
    if path_absolute in relpath:
        return path_absolute
    if relpath.startswith("../"):
        return path_absolute
    return relpath


# That is needed for compatibility with py38, later was added to Path class
def is_relative_to(path: Path, *other: Any) -> bool:
    """Return True if the path is relative to another path or False."""
    try:
        path.resolve().absolute().relative_to(*other)
    except ValueError:
        return False
    return True


def normpath_path(path: str | Path) -> Path:
    """Normalize a path in order to provide a more consistent output.

    - Any symlinks are resolved.
    - Any paths outside the CWD are resolved to their absolute path.
    - Any absolute path within current user home directory is compressed to
      make use of '~', so it is easier to read and more portable.
    """
    if not isinstance(path, Path):
        path = Path(path)

    is_relative = is_relative_to(path, path.cwd())
    path = path.resolve()
    if is_relative:
        path = path.relative_to(path.cwd())

    # Compress any absolute path within current user home directory
    if path.is_absolute():
        home = Path.home()
        if is_relative_to(path, home):
            path = Path("~") / path.relative_to(home)

    return path


@contextmanager
def cwd(path: Path) -> Iterator[None]:
    """Context manager for temporary changing current working directory."""
    old_pwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_pwd)


def expand_path_vars(path: str) -> str:
    """Expand the environment or ~ variables in a path string."""
    # It may be possible for function to be called with a Path object
    path = str(path).strip()
    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    return path


def expand_paths_vars(paths: list[str]) -> list[str]:
    """Expand the environment or ~ variables in a list."""
    paths = [expand_path_vars(p) for p in paths]
    return paths


def kind_from_path(path: Path, *, base: bool = False) -> FileType:
    """Determine the file kind based on its name.

    When called with base=True, it will return the base file type instead
    of the explicit one. That is expected to return 'yaml' for any yaml files.
    """
    # We attempt to use a relative path to the project root for glob matching.
    # This prevents parent directory names (like 'tasks') from triggering
    # false positives in kind discovery. See #4763.
    try:
        project_root, _ = find_project_root([str(path)])
        # .resolve() ensures we handle symlinks and double-dots correctly
        rel_path = path.resolve().relative_to(project_root.resolve())
        pathex = wcmatch.pathlib.PurePath(str(rel_path))
    except (ValueError, RuntimeError):
        # Fallback to absolute if the file is outside the project root or can't be found
        pathex = wcmatch.pathlib.PurePath(str(path.absolute().resolve()))
    kinds = options.kinds if not base else BASE_KINDS
    for entry in kinds:
        for k, v in entry.items():
            if pathex.globmatch(
                v,
                flags=(
                    wcmatch.pathlib.GLOBSTAR
                    | wcmatch.pathlib.BRACE
                    | wcmatch.pathlib.DOTGLOB
                ),
            ):
                matched_kind = str(k)
                # Namespace folders under roles/ can match **/roles/*/ without
                # being role roots themselves. See #5079.
                if (
                    matched_kind == "role"
                    and path.is_dir()
                    and not _has_role_subdirs(path)
                ):
                    continue
                return matched_kind  # type: ignore[return-value]

    if base:
        # Unknown base file type is default
        return ""

    if path.is_dir() and _has_role_subdirs(path):
        return "role"

    if path.is_dir():
        _logger.debug(
            "Folder `%s` does not look like a role due to missing any of the common subfolders such: %s.",
            path,
            ", ".join(ROLE_SUBDIRS),
        )

    if str(path) == "/dev/stdin":
        return "playbook"

    if path.name == "pattern.json":
        # This is a pattern file
        return "pattern"

    # Unknown file types report a empty string (evaluated as False)
    return ""


# pylint: disable=too-many-instance-attributes
class Lintable:
    """Defines a file/folder that can be linted.

    Providing file content when creating the object allow creation of in-memory
    instances that do not need files to be present on disk.

    When symlinks are given, they will always be resolved to their target.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        name: str | Path,
        content: str | None = None,
        kind: FileType | None = None,
        base_kind: str = "",
        parent: Lintable | None = None,
    ):
        """Create a Lintable instance."""
        self.dir: str = ""
        self.kind: FileType | None = None
        self.stop_processing = False  # Set to stop other rules from running
        self.state: Any = States.NOT_LOADED
        self.line_skips: dict[int, set[str]] = defaultdict(set)
        self.exc: Exception | None = None  # Stores data loading exceptions
        self.parent = parent
        self.explicit = False  # Indicates if the file was explicitly provided or was indirectly included.
        self.line_offset = (
            0  # Amount to offset line numbers by to get accurate position
        )
        self.matches: list[MatchError] = []

        if isinstance(name, str):
            name = Path(name)
        is_relative = is_relative_to(name, str(name.cwd()))
        name = name.resolve()
        if is_relative:
            name = name.relative_to(name.cwd())
        name = normpath_path(name)
        # we need to be sure that we expanduser() because otherwise a simple
        # test like .path.exists() will return unexpected results.
        self.path = name.expanduser()
        # Filename is effective file on disk, for stdin is a namedtempfile
        self.name = self.filename = str(name)

        self._content = self._original_content = content
        self.updated = False

        # if the lintable is part of a role, we save role folder name
        self.role = ""
        role_path = find_role_dir(self.path)
        if role_path is not None:
            self.role = role_path.name

        if str(self.path) in ["/dev/stdin", "-"]:
            # pylint: disable=consider-using-with
            self.file = NamedTemporaryFile(  # noqa: SIM115
                mode="w+",
                suffix="playbook.yml",
                encoding="utf-8",
            )
            self.filename = self.file.name
            self._content = sys.stdin.read()
            self.file.write(self._content)
            self.file.flush()
            self.path = Path(self.file.name)
            self.name = "stdin"
            self.kind = "playbook"
            self.dir = "/"
        else:
            self.kind = kind or kind_from_path(self.path)
        # We store absolute directory in dir
        if not self.dir:
            if self.kind == "role":
                self.dir = str(self.path.resolve())
            else:
                self.dir = str(self.path.parent.resolve())

        # determine base file kind (yaml, xml, ini, ...)
        self.base_kind = base_kind or kind_from_path(self.path, base=True)
        self.abspath = self.path.expanduser().absolute()

        if self.kind == "tasks":
            self.parent = _guess_parent(self)

        if self.kind == "yaml":
            _ = self.data

    def __del__(self) -> None:
        """Clean up temporary files when the instance is cleaned up."""
        if hasattr(self, "file"):
            self.file.close()

    def _guess_kind(self) -> None:
        if self.kind == "yaml":
            if (
                isinstance(self.data, list)
                and len(self.data) > 0
                and (
                    "hosts" in self.data[0]
                    or "import_playbook" in self.data[0]
                    or "ansible.builtin.import_playbook" in self.data[0]
                )
            ):
                if "rules" not in self.data[0]:
                    self.kind = "playbook"
                else:
                    self.kind = "rulebook"
            # we we failed to guess the more specific kind, we warn user
            if self.kind == "yaml":
                _logger.debug(
                    "Passed '%s' positional argument was identified as generic '%s' file kind.",
                    self.name,
                    self.kind,
                )

    def __getitem__(self, key: Any) -> Any:
        """Provide compatibility subscriptable support."""
        if key == "path":
            return str(self.path)
        if key == "type":
            return str(self.kind)
        raise NotImplementedError

    def get(self, key: Any, default: Any = None) -> Any:
        """Provide compatibility subscriptable support."""
        try:
            return self[key]
        except NotImplementedError:
            return default

    def _populate_content_cache_from_disk(self) -> None:
        # Can raise UnicodeDecodeError
        self._content = self.path.expanduser().resolve().read_text(encoding="utf-8")

        if self._original_content is None:
            self._original_content = self._content

    @property
    def content(self) -> str:
        """Retrieve file content, from internal cache or disk."""
        if self._content is None:
            self._populate_content_cache_from_disk()
        return cast("str", self._content)

    @content.setter
    def content(self, value: str) -> None:
        """Update ``content`` and calculate ``updated``.

        To calculate ``updated`` this will read the file from disk if the cache
        has not already been populated.
        """
        if not isinstance(value, str):
            msg = f"Expected str but got {type(value)}"
            raise TypeError(msg)
        if self._original_content is None:
            if self._content is not None:
                self._original_content = self._content
            elif self.path.exists():
                self._populate_content_cache_from_disk()
            else:
                # new file
                self._original_content = ""
        self.updated = self._original_content != value
        self._content = value

    @content.deleter
    def content(self) -> None:
        """Reset the internal content cache."""
        self._content = None

    def write(self, *, force: bool = False) -> None:
        """Write the value of ``Lintable.content`` to disk.

        This only writes to disk if the content has been updated (``Lintable.updated``).
        For example, you can update the content, and then write it to disk like this:

        .. code:: python

            lintable.content = new_content
            lintable.write()

        Use ``force=True`` when you want to force a content rewrite even if the
        content has not changed. For example:

        .. code:: python

            lintable.write(force=True)
        """
        dump_filename = self.path.expanduser().resolve()
        if os.environ.get("ANSIBLE_LINT_WRITE_TMP", "0") == "1":
            dump_filename = dump_filename.with_suffix(
                f".tmp{dump_filename.suffix}",
            )
        elif not force and not self.updated:
            # No changes to write.
            return

        dump_filename.write_text(
            self._content or "",
            encoding="utf-8",
        )

    def __hash__(self) -> int:
        """Return a hash value of the lintables."""
        return hash((self.name, self.kind, self.abspath))

    def __eq__(self, other: object) -> bool:
        """Identify whether the other object represents the same rule match."""
        if isinstance(other, Lintable):
            return bool(self.name == other.name and self.kind == other.kind)
        return False

    def __repr__(self) -> str:
        """Return user friendly representation of a lintable."""
        return f"{self.name} ({self.kind})"

    def is_owned_by_ansible(self) -> bool:
        """Return true for YAML files that are managed by Ansible."""
        return self.kind in ANSIBLE_OWNED_KINDS

    def failed(self) -> bool:
        """Return true if we already found syntax-check errors on this file."""
        return any(
            match.rule.id in ("syntax-check", "load-failure", "internal-error")
            for match in self.matches
        )

    def _load_yaml_state(self) -> None:
        """Load and process YAML content, guessing kind and appending skip rules."""
        from ansiblelint.utils import (  # pylint: disable=import-outside-toplevel
            parse_yaml_linenumbers,
        )

        self.state = parse_yaml_linenumbers(self)
        if self.kind == "yaml":
            self._guess_kind()
        if "append_skipped_rules" not in globals():
            # pylint: disable=import-outside-toplevel
            from ansiblelint.skip_utils import append_skipped_rules

        # pylint: disable=possibly-used-before-assignment
        if self.state:
            self.state = append_skipped_rules(self.state, self)

    @property
    def data(self) -> Any:
        """Return loaded data representation for current file, if possible."""
        if self.state == States.NOT_LOADED:
            if self.path.is_dir():
                self.state = None
                return self.state
            try:
                if str(self.base_kind) == "text/yaml":
                    self._load_yaml_state()
                else:
                    _logger.debug(
                        "data set to None for %s due to being '%s' (%s) kind.",
                        self.path,
                        self.kind,
                        self.base_kind or "unknown",
                    )
                    self.state = States.UNKNOWN_DATA

            except (
                RuntimeError,
                FileNotFoundError,
                YAMLError,
                UnicodeDecodeError,
            ) as exc:
                self.state = States.LOAD_FAILED
                self.exc = exc
        return self.state


# pylint: disable=redefined-outer-name
def discover_lintables(options: Options) -> list[str]:
    """Find all files that we know how to lint.

    Return format is normalized, relative for stuff below cwd, ~/ for content
    under current user and absolute for everything else.
    """
    if not options.lintables:
        options.lintables = ["."]

    return [
        str(filename)
        for filename in get_all_files(
            *[Path(s) for s in options.lintables],
            exclude_paths=options.exclude_paths,
        )
    ]


def find_project_root(
    srcs: Sequence[str],
    config_file: str | None = None,
) -> tuple[Path, str]:
    """Return a directory containing .git or ansible-lint config files.

    That directory will be a common parent of all files and directories
    passed in `srcs`.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Returns a two-tuple with the first element as the project root path and
    the second element as a string describing the method by which the
    project root was discovered.
    """
    directory = None
    if not srcs:
        srcs = [str(Path.cwd().resolve().absolute())]
    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    cfg_files = [config_file] if config_file else CONFIG_FILENAMES

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory, ".git directory"

        if (directory / ".hg").is_dir():
            return directory, ".hg directory"

        for cfg_file in cfg_files:
            # note that if cfg_file is already absolute, 'directory' is ignored
            resolved_cfg_path = directory / cfg_file
            if resolved_cfg_path.is_file():
                if os.path.isabs(cfg_file):
                    directory = Path(cfg_file).parent
                    if directory.name == ".config":
                        directory = directory.parent
                return directory, f"config file {resolved_cfg_path}"

    if not directory:
        return Path.cwd(), "current working directory"
    return directory, "file system root"


def path_is_inside(child: Path, parent: Path) -> bool:
    """Return true when child is parent or a descendant of parent."""
    try:
        child.resolve().relative_to(parent.resolve())
    except ValueError:
        return False
    return True


def expand_dirs_in_lintables(lintables: set[Lintable]) -> None:
    """Return all recognized lintables within given directory."""
    should_expand = False

    for item in lintables:
        if item.path.is_dir():
            should_expand = True
            break

    if should_expand:
        # this relies on git and we do not want to call unless needed
        all_files = discover_lintables(options)

        for item in copy.copy(lintables):
            if item.path.is_dir():
                for filename in all_files:
                    if path_is_inside(Path(filename), item.path):
                        lintables.add(Lintable(filename))


def _guess_parent(lintable: Lintable) -> Lintable | None:
    """Return a parent directory for a lintable."""
    role_path = find_role_dir(lintable.path)
    if role_path is not None:
        return Lintable(role_path, kind="role")
    return None


def get_all_files(
    *paths: Path,
    exclude_paths: list[str] | None = None,
) -> list[Path]:
    """Recursively retrieve all files from given folders."""
    all_files: list[Path] = []
    exclude_paths = [] if exclude_paths is None else exclude_paths

    def is_excluded(path_to_check: Path) -> bool:
        """Check if a file is exclude by current specs."""
        return any(
            spec.match_file(pathspec.util.append_dir_sep(path_to_check))
            for spec in pathspecs
        )

    for path in paths:
        pathspecs = [
            pathspec.GitIgnoreSpec.from_lines(
                [
                    ".ansible",
                    ".git",
                    ".tox",
                    ".mypy_cache",
                    "__pycache__",
                    ".DS_Store",
                    ".coverage",
                    ".pytest_cache",
                    ".ruff_cache",
                    *exclude_paths,
                ],
            ),
        ]
        gitignore = path / ".gitignore"
        if gitignore.exists():
            with gitignore.open(encoding="UTF-8") as f:
                _logger.info("Loading ignores from %s", gitignore)
                pathspecs.append(
                    pathspec.GitIgnoreSpec.from_lines(f.read().splitlines()),
                )

        # Iterate over all items in the directory
        if path.is_file():
            all_files.append(path)
        else:
            for item in sorted(path.iterdir()):
                if is_excluded(item):
                    _logger.debug("Excluded: %s", item)
                    continue
                if item.is_file():
                    all_files.append(item)
                # If it's a directory, recursively call the function
                elif item.is_dir():
                    all_files.extend(get_all_files(item, exclude_paths=exclude_paths))

    return all_files



"""Module containing cached JSON schemas."""

from __future__ import annotations

import json
import logging
import re
from typing import TYPE_CHECKING, Any

import yaml
from jsonschema import Draft202012Validator
from jsonschema.exceptions import ValidationError

from ansiblelint.loaders import yaml_load_safe
from ansiblelint.schemas.__main__ import JSON_SCHEMAS, _schema_cache

_logger = logging.getLogger(__package__)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from ansiblelint.file_utils import Lintable


RE_MD_URLS = re.compile(
    r"\[.{0,256}?\]\((?P<url>https?://[^\s]+)\)|(?P<url2>https?://[^\s]+)"
)


def find_best_deep_match(
    errors: ValidationError,
) -> ValidationError:
    """Return the deepest schema validation error."""

    def iter_validation_error(
        err: ValidationError,
    ) -> Iterator[ValidationError]:
        if err.context:
            for e in err.context:
                yield e
                yield from iter_validation_error(e)

    return max(iter_validation_error(errors), key=_deep_match_relevance)


def _find_documentation_url(*schemas: dict[Any, Any]) -> str:
    """Search schemas for a documentation URL."""
    for json_schema in schemas:
        for k in ("description", "markdownDescription"):
            if k in json_schema:
                match = RE_MD_URLS.search(json_schema[k])
                if match:
                    return next(x for x in match.groups() if x is not None)
    return ""


def _format_validation_message(
    error: ValidationError,
    schema: dict[Any, Any],
) -> str:
    """Format a validation error into a user-facing message."""
    if not hasattr(error, "schema") or not isinstance(
        error.schema, dict
    ):  # pragma: no cover
        msg = "error object does not have schema attribute"
        raise TypeError(msg)
    if "not" in error.schema and len(error.schema["not"]) == 0:
        message: str = error.schema["title"]
        schema = error.schema
    else:
        message = f"{error.json_path} {error.message}"

    documentation_url = _find_documentation_url(error.schema, schema)
    if documentation_url:
        if not message.endswith("."):
            message += "."
        message += f" See {documentation_url}"
    return message


def _validate_json_data(
    json_data: Any,
    schema: dict[Any, Any],
) -> str | None:
    """Validate JSON data against schema, return message or None if valid."""
    validator = Draft202012Validator(schema)
    try:
        error = next(validator.iter_errors(json_data))
    except StopIteration:
        return None
    if error.context:
        error = find_best_deep_match(error)
    return _format_validation_message(error, schema)


def validate_file_schema(file: Lintable) -> list[str]:
    """Return list of JSON validation errors found."""
    schema: dict[Any, Any] = {}
    if file.kind not in JSON_SCHEMAS:
        return [f"Unable to find JSON Schema '{file.kind}' for '{file.path}' file."]
    try:
        yaml_data = yaml_load_safe(file.content)
        json_data = json.loads(json.dumps(yaml_data))
        schema = _schema_cache[file.kind]
        message = _validate_json_data(json_data, schema)
        if message is None:
            return []
    except yaml.constructor.ConstructorError as exc:
        return [f"Failed to load YAML file '{file.path}': {exc.problem}"]
    except ValidationError as exc:  # pragma: no cover
        message = exc.message
        documentation_url = _find_documentation_url(schema)
        if documentation_url:
            if not message.endswith("."):
                message += "."
            message += f" See {documentation_url}"
        return [message]
    return [message]


def _deep_match_relevance(error: ValidationError) -> tuple[bool | int, ...]:
    validator = error.validator
    return (
        validator not in ("anyOf", "oneOf"),
        len(error.absolute_path),
        -len(error.path),
    )



"""Implementation of no-tabs rule."""

# Copyright (c) 2016, Will Thames and contributors
# Copyright (c) 2018, Ansible Project
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ansiblelint.rules import AnsibleLintRule
from ansiblelint.text import has_jinja
from ansiblelint.yaml_utils import nested_items_path

if TYPE_CHECKING:
    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class NoTabsRule(AnsibleLintRule):
    """Most files should not contain tabs."""

    id = "no-tabs"
    description = "Tabs can cause unexpected display issues, use spaces"
    severity = "LOW"
    tags = ["formatting"]
    version_changed = "4.0.0"
    allow_list = [
        ("lineinfile", "insertafter"),
        ("lineinfile", "insertbefore"),
        ("lineinfile", "regexp"),
        ("lineinfile", "line"),
        ("win_lineinfile", "insertafter"),
        ("win_lineinfile", "insertbefore"),
        ("win_lineinfile", "regexp"),
        ("win_lineinfile", "line"),
        ("ansible.builtin.lineinfile", "insertafter"),
        ("ansible.builtin.lineinfile", "insertbefore"),
        ("ansible.builtin.lineinfile", "regexp"),
        ("ansible.builtin.lineinfile", "line"),
        ("ansible.legacy.lineinfile", "insertafter"),
        ("ansible.legacy.lineinfile", "insertbefore"),
        ("ansible.legacy.lineinfile", "regexp"),
        ("ansible.legacy.lineinfile", "line"),
        ("community.windows.win_lineinfile", "insertafter"),
        ("community.windows.win_lineinfile", "insertbefore"),
        ("community.windows.win_lineinfile", "regexp"),
        ("community.windows.win_lineinfile", "line"),
    ]

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> list[MatchError]:
        result = []
        # Check the key/value pairs found by the nested pathing
        for k, v, _path in nested_items_path(task):
            # Check if the Key itself has a tab (almost never allowed)
            if isinstance(k, str) and "\t" in k and not has_jinja(k):
                result.append(
                    self.create_matcherror(
                        message=self.shortdesc,
                        data=k,
                        filename=file,
                    )
                )

            # Check if the Value has a tab
            if isinstance(v, str) and "\t" in v and not has_jinja(v):
                # We check if 'k' is in our allow_list for ANY of the modules
                is_allowed = any(k == allowed_key for _, allowed_key in self.allow_list)

                if not is_allowed:
                    result.append(
                        self.create_matcherror(
                            message=self.shortdesc,
                            data=v,
                            filename=file,
                        )
                    )
        return result


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.libyaml
    def test_no_tabs_rule(default_rules_collection: RulesCollection) -> None:
        """Test rule matches."""
        results = Runner(
            "examples/playbooks/rule-no-tabs.yml",
            rules=default_rules_collection,
        ).run()
        lines = []
        for result in results:
            assert result.rule.id == "no-tabs"
            lines.append(result.lineno)
        assert lines
        # 2.19 has more precise line:columns numbers so the effective result
        # is different.
        assert lines == [10, 13] or lines == [12, 15, 15], lines

    @pytest.mark.libyaml
    def test_no_tabs_block_pass(default_rules_collection: RulesCollection) -> None:
        """Verify that tabs are allowed in lineinfile even inside blocks."""
        results = Runner(
            "examples/playbooks/rule-no-tabs-block-pass.yml",
            rules=default_rules_collection,
        ).run()
        assert len(results) == 0



"""Rule definition for JSON Schema Validations."""

from __future__ import annotations

import logging
import re
import sys
from typing import TYPE_CHECKING, Any

from ansiblelint.file_utils import Lintable
from ansiblelint.rules import AnsibleLintRule
from ansiblelint.schemas.__main__ import JSON_SCHEMAS
from ansiblelint.schemas.main import validate_file_schema
from ansiblelint.text import has_jinja

if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from ansiblelint.app import App
    from ansiblelint.config import Options
    from ansiblelint.errors import MatchError
    from ansiblelint.utils import Task


_logger = logging.getLogger(__name__)


DESCRIPTION_MD = """ Returned errors will not include exact line numbers, but they will mention
the schema name being used as a tag, like ``schema[playbook]``,
``schema[tasks]``.

This rule is not skippable and stops further processing of the file.

If incorrect schema was picked, you might want to either:

* move the file to standard location, so its file is detected correctly.
* use ``kinds:`` option in linter config to help it pick correct file type.
"""

pre_checks = {
    "task": {
        "with_flattened": {
            "msg": "with_flattened was moved to with_community.general.flattened in 2.10",
            "tag": "moves",
        },
        "with_filetree": {
            "msg": "with_filetree was moved to with_community.general.filetree in 2.10",
            "tag": "moves",
        },
        "with_cartesian": {
            "msg": "with_cartesian was moved to with_community.general.flattened in 2.10",
            "tag": "moves",
        },
    },
}


class ValidateSchemaRule(AnsibleLintRule):
    """Perform JSON Schema Validation for known lintable kinds."""

    description = DESCRIPTION_MD

    id = "schema"
    severity = "VERY_HIGH"
    tags = ["core"]
    version_changed = "25.9.0"
    _ids = {
        "schema[ansible-lint-config]": "",
        "schema[ansible-navigator-config]": "",
        "schema[changelog]": "",
        "schema[execution-environment]": "",
        "schema[galaxy]": "",
        "schema[inventory]": "",
        "schema[meta]": "",
        "schema[meta-runtime]": "",
        "schema[molecule]": "",
        "schema[play-argspec]": "",
        "schema[playbook]": "",
        "schema[requirements]": "",
        "schema[role-arg-spec]": "",
        "schema[rulebook]": "",
        "schema[tasks]": "",
        "schema[vars]": "",
    }
    _field_checks: dict[str, list[str]] = {}

    @property
    def field_checks(self) -> dict[str, list[str]]:
        """Lazy property for returning field checks."""
        if not self._collection:
            msg = "Rule was not registered to a RuleCollection."
            raise RuntimeError(msg)
        if not self._field_checks:
            self._field_checks = {
                "become_method": sorted(
                    self._collection.app.runtime.plugins.become.keys(),
                ),
            }
        return self._field_checks

    def matchplay(self, file: Lintable, data: dict[str, Any]) -> list[MatchError]:
        """Return matches found for a specific playbook."""
        results: list[MatchError] = []
        if (
            not data
            or file.kind not in ("tasks", "handlers", "playbook")
            or file.failed()
        ):
            return results
        # check at play level
        results.extend(self._get_field_matches(file=file, data=data))
        return results

    def _get_field_matches(
        self,
        file: Lintable | None,
        data: MutableMapping[str, Any],
    ) -> list[MatchError]:
        """Retrieve all matches related to fields for the given data block."""
        results = []
        kind = "tasks" if not file else file.kind
        for key, values in self.field_checks.items():
            if key in data:
                plugin_value = data[key]
                if not has_jinja(plugin_value) and plugin_value not in values:
                    msg = f"'{key}' must be one of the currently available values: {', '.join(values)}"
                    results.append(
                        self.create_matcherror(
                            message=msg,
                            data=plugin_value,
                            filename=file,
                            details=ValidateSchemaRule.description,
                            tag=f"schema[{kind}]",
                        ),
                    )
        return results

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> bool | str | MatchError | list[MatchError]:
        results: list[MatchError] = []
        if not file:
            file = Lintable("", kind="tasks")

        if file and file.failed():
            return results

        results.extend(self._get_field_matches(file=file, data=task.raw_task))
        for key in pre_checks["task"]:
            if key in task.raw_task:
                msg = pre_checks["task"][key]["msg"]
                tag = pre_checks["task"][key]["tag"]
                results.append(
                    self.create_matcherror(
                        message=msg,
                        filename=file,
                        details=ValidateSchemaRule.description,
                        tag=f"schema[{tag}]",
                    ),
                )
        return results

    def matchyaml(self, file: Lintable) -> list[MatchError]:
        """Return JSON validation errors found as a list of MatchError(s)."""
        result: list[MatchError] = []

        if file.failed():
            return result

        if file.kind not in JSON_SCHEMAS:
            return result

        for error in validate_file_schema(file):
            if error.startswith("Failed to load YAML file"):
                _logger.debug(
                    "Ignored failure to load %s for schema validation, as !vault may cause it.",
                    file,
                )
                return []

            result.append(
                self.create_matcherror(
                    message=error,
                    filename=file,
                    details=ValidateSchemaRule.description,
                    tag=f"schema[{file.kind}]",
                ),
            )
            break

        if not result:
            result = super().matchyaml(file)
        return result


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("file", "expected_kind", "expected"),
        (
            pytest.param(
                "examples/.collection/galaxy.yml",
                "galaxy",
                [r".*'GPL' is not one of.*https://"],
                id="galaxy",
            ),
            pytest.param(
                "examples/roles/invalid_requirements_schema/meta/requirements.yml",
                "requirements",
                [
                    # r".*{'foo': 'bar'} is not valid under any of the given schemas.*https://",
                    r".*{'foo': 'bar'} is not of type 'array'.*https://",
                ],
                id="requirements",
            ),
            pytest.param(
                "examples/roles/invalid_meta_schema/meta/main.yml",
                "meta",
                [r".*False is not of type 'string'.*https://"],
                id="meta",
            ),
            pytest.param(
                "examples/playbooks/vars/invalid_vars_schema.yml",
                "vars",
                [r".* '123' does not match any of the regexes.*https://"],
                id="vars",
            ),
            pytest.param(
                "examples/execution-environment.yml",
                "execution-environment",
                [],
                id="execution-environment",
            ),
            pytest.param(
                "examples/ee_broken/execution-environment.yml",
                "execution-environment",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="execution-environment-broken",
            ),
            pytest.param(
                "examples/meta/runtime.yml",
                "meta-runtime",
                [],
                id="meta-runtime",
            ),
            pytest.param(
                "examples/broken_collection_meta_runtime/meta/runtime.yml",
                "meta-runtime",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="meta-runtime-broken",
            ),
            pytest.param(
                "examples/inventory/production.yml",
                "inventory",
                [],
                id="inventory",
            ),
            pytest.param(
                "examples/inventory/broken_dev_inventory.yml",
                "inventory",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="inventory-broken",
            ),
            pytest.param(
                ".ansible-lint",
                "ansible-lint-config",
                [],
                id="ansible-lint-config",
            ),
            pytest.param(
                "examples/.config/ansible-lint.yml",
                "ansible-lint-config",
                [],
                id="ansible-lint-config2",
            ),
            pytest.param(
                "examples/broken/.ansible-lint",
                "ansible-lint-config",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="ansible-lint-config-broken",
            ),
            pytest.param(
                "examples/broken_supported_ansible_also/.ansible-lint",
                "ansible-lint-config",
                [
                    r".*supported_ansible_also True is not of type 'array'.*https://",
                ],
                id="ansible-lint-config-broken",
            ),
            pytest.param(
                "examples/ansible-navigator.yml",
                "ansible-navigator-config",
                [],
                id="ansible-navigator-config",
            ),
            pytest.param(
                "examples/broken/ansible-navigator.yml",
                "ansible-navigator-config",
                [
                    r".*Additional properties are not allowed \('ansible' was unexpected\).*https://",
                ],
                id="ansible-navigator-config-broken",
            ),
            pytest.param(
                "examples/roles/hello/meta/argument_specs.yml",
                "role-arg-spec",
                [],
                id="role-arg-spec",
            ),
            pytest.param(
                "examples/roles/broken_argument_specs/meta/argument_specs.yml",
                "role-arg-spec",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="role-arg-spec-broken",
            ),
            pytest.param(
                "examples/changelogs/changelog.yaml",
                "changelog",
                [
                    r".*Additional properties are not allowed \('foo' was unexpected\).*https://",
                ],
                id="changelog",
            ),
            pytest.param(
                "examples/rulebooks/rulebook-fail.yml",
                "rulebook",
                [
                    # r".*Additional properties are not allowed \('that_should_not_be_here' was unexpected\).*https://",
                    r".*'sss' is not of type 'object'.*https://",
                ],
                id="rulebook",
            ),
            pytest.param(
                "examples/rulebooks/rulebook-pass.yml",
                "rulebook",
                [],
                id="rulebook2",
            ),
            pytest.param(
                "examples/playbooks/rule-schema-become-method-pass.yml",
                "playbook",
                [],
                id="playbook",
            ),
            pytest.param(
                "examples/playbooks/rule-schema-become-method-fail.yml",
                "playbook",
                [
                    "'become_method' must be one of the currently available values",
                    "'become_method' must be one of the currently available values",
                ],
                id="playbook2",
            ),
            pytest.param(
                "examples/play_argspecs/correct_play_argspec/patterns/example_pattern/playbooks/site.meta.yml",
                "play-argspec",
                [],
                id="play_argspec_positive",
            ),
            pytest.param(
                "examples/play_argspecs/incorrect_play_argspec/site.meta.yaml",
                "play-argspec",
                [
                    r"\$.argument_specs.weather_check 'options' is a required property.",
                ],
                id="play_argspec_negative",
            ),
        ),
    )
    def test_schema(
        file: str,
        expected_kind: str,
        expected: list[str],
        config_options: Options,
        app: App,
    ) -> None:
        """Validate parsing of ansible output."""
        lintable = Lintable(file)
        assert lintable.kind == expected_kind

        rules = RulesCollection(app=app, options=config_options)
        rules.register(ValidateSchemaRule())
        results = Runner(lintable, rules=rules).run()

        assert len(results) == len(expected), results
        for idx, result in enumerate(results):
            assert result.filename.endswith(file)
            assert re.match(expected[idx], result.message)
            assert result.tag == f"schema[{expected_kind}]"

    @pytest.mark.parametrize(
        ("file", "expected_kind", "expected_tag", "count"),
        (
            pytest.param(
                "examples/playbooks/rule-syntax-moves.yml",
                "playbook",
                "schema[moves]",
                3,
                id="playbook",
            ),
        ),
    )
    def test_schema_moves(
        file: str,
        expected_kind: str,
        expected_tag: str,
        count: int,
        config_options: Options,
        app: App,
    ) -> None:
        """Validate ability to detect schema[moves]."""
        lintable = Lintable(file)
        assert lintable.kind == expected_kind

        rules = RulesCollection(app=app, options=config_options)
        rules.register(ValidateSchemaRule())
        results = Runner(lintable, rules=rules).run()

        assert len(results) == count, results
        for result in results:
            assert result.filename.endswith(file)
            assert result.tag == expected_tag



"""Optional Ansible-lint rule to enforce use of prefix on role loop vars."""

from __future__ import annotations

import re
import sys
from typing import TYPE_CHECKING

from ansiblelint.config import LOOP_VAR_PREFIX, options
from ansiblelint.rules import AnsibleLintRule
from ansiblelint.text import toidentifier

if TYPE_CHECKING:
    from ansiblelint.errors import MatchError
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class RoleLoopVarPrefix(AnsibleLintRule):
    """Role loop_var should use configured prefix."""

    id = "loop-var-prefix"
    link = (
        "https://docs.ansible.com/projects/ansible/latest/playbook_guide/"
        "playbooks_loops.html#defining-inner-and-outer-variable-names-with-loop-var"
    )
    description = """\
Looping inside roles has the risk of clashing with loops from user-playbooks.\
"""

    tags = ["idiom"]
    prefix = re.compile(r"")
    severity = "MEDIUM"
    _ids = {
        "loop-var-prefix[wrong]": "Loop variable name does not match regex.",
        "loop-var-prefix[missing]": "Replace unsafe implicit `item` loop variable.",
    }
    version_changed = "6.7.0"

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> list[MatchError]:
        """Return matches for a task."""
        if not file or not file.role or not options.loop_var_prefix:
            return []

        self.prefix = re.compile(
            options.loop_var_prefix.format(role=toidentifier(file.role)),
        )
        has_loop = False
        if "loop" in task.raw_task:
            data = task.raw_task["loop"]
            has_loop = True
        else:
            for key in task.raw_task:
                if key.startswith("with_"):
                    data = key
                    has_loop = True
                    break

        if has_loop:
            loop_control = task.raw_task.get("loop_control", {})
            loop_var = loop_control.get("loop_var", "")

            if loop_var:
                if not self.prefix.match(loop_var):
                    return [
                        self.create_matcherror(
                            message=f"Loop variable name does not match /{options.loop_var_prefix}/ regex, where role={toidentifier(file.role)}.",
                            filename=file,
                            data=loop_var,
                            tag="loop-var-prefix[wrong]",
                        ),
                    ]
            else:
                return [
                    self.create_matcherror(
                        message=f"Replace unsafe implicit `item` loop variable by adding a `loop_var` that is matching /{options.loop_var_prefix}/ regex.",
                        filename=file,
                        data=data,
                        tag="loop-var-prefix[missing]",
                    ),
                ]

        return []


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("test_file", "failures"),
        (
            pytest.param(
                "examples/playbooks/roles/loop_var_prefix/tasks/pass.yml",
                0,
                id="pass",
            ),
            pytest.param(
                "examples/playbooks/roles/loop_var_prefix/tasks/fail.yml",
                6,
                id="fail",
            ),
        ),
    )
    def test_loop_var_prefix(
        default_rules_collection: RulesCollection,
        test_file: str,
        failures: int,
    ) -> None:
        """Test rule matches."""
        # Enable checking of loop variable prefixes in roles
        options.loop_var_prefix = LOOP_VAR_PREFIX
        results = Runner(test_file, rules=default_rules_collection).run()
        for result in results:
            assert result.rule.id == RoleLoopVarPrefix().id
        assert len(results) == failures



"""Ansible-lint version information."""

# this is the fallback SemVer version picked by setuptools_scm when tag
# information is not available.
__version__ = "0.1.dev1"
__all__ = ("__version__",)
from typing import TYPE_CHECKING

# as either pyright or mypy have problems with these import fallbacks, we
# avoid running them when doing type checking
if not TYPE_CHECKING:
    try:
        from ._version import version as __version__
    except ImportError:  # pragma: no cover
        try:
            import pkg_resources  # pylint: disable=import-error

            __version__ = pkg_resources.get_distribution("ansible-lint").version
        except Exception:  # pylint: disable=broad-except  # noqa: BLE001, S110
            pass



"""IgnoreErrorsRule used with ansible-lint."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ansiblelint.rules import AnsibleLintRule

if TYPE_CHECKING:
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class IgnoreErrorsRule(AnsibleLintRule):
    """Use failed_when and specify error conditions instead of using ignore_errors."""

    id = "ignore-errors"
    description = (
        "Instead of ignoring all errors, ignore the errors only when using ``{{ ansible_check_mode }}``, "
        "register the errors using ``register``, "
        "or use ``failed_when:`` and specify acceptable error conditions "
        "to reduce the risk of ignoring important failures."
    )
    severity = "LOW"
    tags = ["unpredictability"]
    version_changed = "5.0.7"

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> bool | str:
        return (
            task.get("ignore_errors")
            and task.get("ignore_errors") != "{{ ansible_check_mode }}"
            and not task.get("register")
        )


if "pytest" in sys.modules:
    import pytest

    if TYPE_CHECKING:
        from ansiblelint.testing import RunFromText

    IGNORE_ERRORS_TRUE = """
- hosts: all
  tasks:
    - name: Run apt-get update
      command: apt-get update
      ignore_errors: true
"""

    IGNORE_ERRORS_FALSE = """
- hosts: all
  tasks:
    - name: Run apt-get update
      command: apt-get update
      ignore_errors: false
"""

    IGNORE_ERRORS_CHECK_MODE = """
- hosts: all
  tasks:
    - name: Run apt-get update
      command: apt-get update
      ignore_errors: "{{ ansible_check_mode }}"
"""

    IGNORE_ERRORS_REGISTER = """
- hosts: all
  tasks:
    - name: Run apt-get update
      command: apt-get update
      ignore_errors: true
      register: ignore_errors_register
"""

    FAILED_WHEN = """
- hosts: all
  tasks:
    - name: Disable apport
      become: 'yes'
      lineinfile:
        line: "enabled=0"
        dest: /etc/default/apport
        mode: 0644
        state: present
      register: default_apport
      failed_when: default_apport.rc !=0 and not default_apport.rc == 257
"""

    @pytest.mark.parametrize(
        "rule_runner",
        (IgnoreErrorsRule,),
        indirect=["rule_runner"],
    )
    def test_ignore_errors_true(rule_runner: RunFromText) -> None:
        """The task uses ignore_errors."""
        results = rule_runner.run_playbook(IGNORE_ERRORS_TRUE)
        assert len(results) == 1

    @pytest.mark.parametrize(
        "rule_runner",
        (IgnoreErrorsRule,),
        indirect=["rule_runner"],
    )
    def test_ignore_errors_false(rule_runner: RunFromText) -> None:
        """The task uses ignore_errors: false, oddly enough."""
        results = rule_runner.run_playbook(IGNORE_ERRORS_FALSE)
        assert len(results) == 0

    @pytest.mark.parametrize(
        "rule_runner",
        (IgnoreErrorsRule,),
        indirect=["rule_runner"],
    )
    def test_ignore_errors_check_mode(rule_runner: RunFromText) -> None:
        """The task uses ignore_errors: "{{ ansible_check_mode }}"."""
        results = rule_runner.run_playbook(IGNORE_ERRORS_CHECK_MODE)
        assert len(results) == 0

    @pytest.mark.parametrize(
        "rule_runner",
        (IgnoreErrorsRule,),
        indirect=["rule_runner"],
    )
    def test_ignore_errors_register(rule_runner: RunFromText) -> None:
        """The task uses ignore_errors: but output is registered and managed."""
        results = rule_runner.run_playbook(IGNORE_ERRORS_REGISTER)
        assert len(results) == 0

    @pytest.mark.parametrize(
        "rule_runner",
        (IgnoreErrorsRule,),
        indirect=["rule_runner"],
    )
    def test_failed_when(rule_runner: RunFromText) -> None:
        """Instead of ignore_errors, this task uses failed_when."""
        results = rule_runner.run_playbook(FAILED_WHEN)
        assert len(results) == 0



"""Implementation of no-relative-paths rule."""

# Copyright (c) 2016, Tsukinowa Inc. <info@tsukinowa.jp>
# Copyright (c) 2018, Ansible Project

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ansiblelint.rules import AnsibleLintRule

if TYPE_CHECKING:
    from ansiblelint.file_utils import Lintable
    from ansiblelint.utils import Task


class RoleRelativePath(AnsibleLintRule):
    """The src argument should not use a relative path."""

    id = "no-relative-paths"
    description = "The ``copy`` and ``template`` modules should not use relative path for ``src``."
    severity = "HIGH"
    tags = ["idiom"]
    version_changed = "4.0.0"

    _module_to_path_folder = {
        "copy": "files",
        "win_copy": "files",
        "template": "templates",
        "win_template": "win_templates",
    }

    def matchtask(
        self,
        task: Task,
        file: Lintable | None = None,
    ) -> bool | str:
        module = task["action"]["__ansible_module__"]
        if module not in self._module_to_path_folder:
            return False

        if "src" not in task["action"]:
            return False

        path_to_check = f"../{self._module_to_path_folder[module]}"
        return path_to_check in task["action"]["src"]


# testing code to be loaded only with pytest or when executed the rule file
if "pytest" in sys.modules:
    import pytest

    # pylint: disable=ungrouped-imports
    from ansiblelint.rules import RulesCollection
    from ansiblelint.runner import Runner

    @pytest.mark.parametrize(
        ("test_file", "failures"),
        (
            pytest.param("examples/playbooks/no_relative_paths_fail.yml", 2, id="fail"),
            pytest.param("examples/playbooks/no_relative_paths_pass.yml", 0, id="pass"),
        ),
    )
    def test_no_relative_paths(
        default_rules_collection: RulesCollection,
        test_file: str,
        failures: int,
    ) -> None:
        """Test rule matches."""
        results = Runner(test_file, rules=default_rules_collection).run()
        assert len(results) == failures
        for result in results:
            assert result.tag == "no-relative-paths"
